@@ -35,7 +35,6 @@ from mistral.db.sqlalchemy import sqlite_lock
 from mistral.db import utils as m_dbutils
 from mistral.db.v2.sqlalchemy import filters as db_filters
 from mistral.db.v2.sqlalchemy import models
-from mistral.db.v2.sqlalchemy.models import WorkflowExecution
 from mistral import exceptions as exc
 from mistral.services import security
 from mistral import utils
@@ -222,13 +221,7 @@ def _get_collection(model, insecure=False, limit=None, marker=None,
         query
     )
 
-    try:
-        return query.all()
-    except Exception as e:
-        raise exc.DBQueryEntryError(
-            "Failed when querying database, error type: %s, "
-            "error message: %s" % (e.__class__.__name__, str(e))
-        )
+    return query.all()
 
 
 def _get_collection_sorted_by_name(model, insecure=False, fields=None,
@@ -273,11 +266,7 @@ def _get_wf_object_by_name_and_namespace(model, name, namespace=None):
     filter_ = model.name == name
 
     if namespace is not None:
-        in_namespace = sa.or_(
-            model.namespace == namespace,
-            model.namespace == ''
-        )
-        filter_ = sa.and_(filter_, in_namespace)
+        filter_ = sa.and_(filter_, model.namespace.in_([namespace, '']))
 
         # Give priority to objects not in the default namespace.
         query = query.order_by(model.namespace.desc())
@@ -508,10 +497,7 @@ def get_workflow_definition_by_id(id, session=None):
 def load_workflow_definition(name, namespace='', session=None):
     model = models.WorkflowDefinition
 
-    filter_ = sa.or_(
-        model.namespace == namespace,
-        model.namespace == ''
-    )
+    filter_ = model.namespace.in_([namespace, ''])
 
     # Give priority to objects not in the default namespace.
     order_by = model.namespace.desc()
@@ -883,6 +869,34 @@ def delete_workflow_executions(session=None, **kwargs):
     return _delete_all(models.WorkflowExecution, **kwargs)
 
 
+@b.session_aware()
+def update_workflow_execution_state(id, cur_state, state, session=None):
+    wf_ex = None
+
+    # Use WHERE clause to exclude possible conflicts if the state has
+    # already been changed.
+    try:
+        specimen = models.WorkflowExecution(
+            id=id,
+            state=cur_state
+        )
+
+        wf_ex = b.model_query(
+            models.WorkflowExecution).update_on_match(
+            specimen=specimen,
+            surrogate_key='id',
+            values={'state': state}
+        )
+    except oslo_sqlalchemy.update_match.NoRowsMatched:
+        LOG.info(
+            "Can't change workflow execution state from %s to %s, "
+            "because it has already been changed. [execution_id=%s]",
+            cur_state, state, id
+        )
+
+    return wf_ex
+
+
 # Tasks executions.
 
 @b.session_aware()
@@ -913,10 +927,10 @@ def _get_completed_task_executions_query(kwargs):
     query = query.filter_by(**kwargs)
 
     query = query.filter(
-        sa.or_(
-            models.TaskExecution.state == states.ERROR,
-            models.TaskExecution.state == states.CANCELLED,
-            models.TaskExecution.state == states.SUCCESS
+        models.TaskExecution.state.in_(
+            [states.ERROR,
+             states.CANCELLED,
+             states.SUCCESS]
         )
     )
 
@@ -936,12 +950,12 @@ def _get_incomplete_task_executions_query(kwargs):
     query = query.filter_by(**kwargs)
 
     query = query.filter(
-        sa.or_(
-            models.TaskExecution.state == states.IDLE,
-            models.TaskExecution.state == states.RUNNING,
-            models.TaskExecution.state == states.WAITING,
-            models.TaskExecution.state == states.RUNNING_DELAYED,
-            models.TaskExecution.state == states.PAUSED
+        models.TaskExecution.state.in_(
+            [states.IDLE,
+             states.RUNNING,
+             states.WAITING,
+             states.RUNNING_DELAYED,
+             states.PAUSED]
         )
     )
 
@@ -1011,6 +1025,34 @@ def _get_task_executions(**kwargs):
     return _get_collection_sorted_by_time(models.TaskExecution, **kwargs)
 
 
+@b.session_aware()
+def update_task_execution_state(id, cur_state, state, session=None):
+    wf_ex = None
+
+    # Use WHERE clause to exclude possible conflicts if the state has
+    # already been changed.
+    try:
+        specimen = models.TaskExecution(
+            id=id,
+            state=cur_state
+        )
+
+        wf_ex = b.model_query(
+            models.TaskExecution).update_on_match(
+            specimen=specimen,
+            surrogate_key='id',
+            values={'state': state}
+        )
+    except oslo_sqlalchemy.update_match.NoRowsMatched:
+        LOG.info(
+            "Can't change task execution state from %s to %s, "
+            "because it has already been changed. [execution_id=%s]",
+            cur_state, state, id
+        )
+
+    return wf_ex
+
+
 # Delayed calls.
 
 @b.session_aware()
@@ -1099,30 +1141,31 @@ def delete_delayed_calls(session=None, **kwargs):
 
 
 @b.session_aware()
-def get_executions_to_clean(expiration_time, limit=None,
-                            max_finished_executions=None, columns=(),
-                            session=None):
-    # Get the ids of the executions that won't be deleted.
-    # These are the not expired executions,
-    # limited by the new max_finished_executions constraint.
-    query = _get_completed_root_executions_query((WorkflowExecution.id,))
-    query = query.filter(
-        models.WorkflowExecution.updated_at >= expiration_time
-    )
-    query = query.order_by(models.WorkflowExecution.updated_at.desc())
-
-    if max_finished_executions:
-        query = query.limit(max_finished_executions)
-
-    # And take the inverse of that set.
-    inverse = _get_completed_root_executions_query(columns)
-    inverse = inverse.filter(~WorkflowExecution.id.in_(query))
-    inverse = inverse.order_by(models.WorkflowExecution.updated_at.asc())
+def get_expired_executions(expiration_time, limit=None, columns=(),
+                           session=None):
+    query = _get_completed_root_executions_query(columns)
+    query = query.filter(models.WorkflowExecution.updated_at < expiration_time)
 
     if limit:
-        inverse.limit(limit)
+        query = query.limit(limit)
 
-    return inverse.all()
+    return query.all()
+
+
+@b.session_aware()
+def get_superfluous_executions(max_finished_executions, limit=None, columns=(),
+                               session=None):
+    if not max_finished_executions:
+        return []
+
+    query = _get_completed_root_executions_query(columns)
+    query = query.order_by(models.WorkflowExecution.updated_at.desc())
+    query = query.offset(max_finished_executions)
+
+    if limit:
+        query = query.limit(limit)
+
+    return query.all()
 
 
 def _get_completed_root_executions_query(columns):
@@ -1131,10 +1174,10 @@ def _get_completed_root_executions_query(columns):
     query = query.filter(models.WorkflowExecution.
                          task_execution_id == sa.null())
     query = query.filter(
-        sa.or_(
-            models.WorkflowExecution.state == states.SUCCESS,
-            models.WorkflowExecution.state == states.ERROR,
-            models.WorkflowExecution.state == states.CANCELLED
+        models.WorkflowExecution.state.in_(
+            [states.SUCCESS,
+             states.ERROR,
+             states.CANCELLED]
         )
     )
     return query
@@ -1621,10 +1664,14 @@ def delete_named_lock(lock_id, session=None):
 
 @contextlib.contextmanager
 def named_lock(name):
-    lock_id = None
+    # NOTE(rakhmerov): We can't use the well-known try-finally pattern here
+    # because if lock creation failed then it means that the SQLAlchemy
+    # session is no longer valid and we can't use to try to delete the lock.
+    # All we can do here is to let the exception bubble up so that the
+    # transaction management code could rollback the transaction.
 
-    try:
-        lock_id = create_named_lock(name)
-        yield
-    finally:
-        delete_named_lock(lock_id)
+    lock_id = create_named_lock(name)
+
+    yield
+
+    delete_named_lock(lock_id)
