@@ -11,7 +11,12 @@
 #    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 #    See the License for the specific language governing permissions and
 #    limitations under the License.
+import socket
 
+import itertools
+
+import errno
+import six
 from six import moves
 
 import kombu
@@ -25,6 +30,11 @@ from mistral.rpc.kombu import kombu_hosts
 from mistral.rpc.kombu import kombu_listener
 from mistral import utils
 
+#: When connection to the RabbitMQ server breaks, the
+#: client will receive EPIPE socket errors. These indicate
+#: an error that may be fixed by retrying. This constant
+#: is a guess for how many times the retry may be reasonable
+EPIPE_RETRIES = 4
 
 LOG = logging.getLogger(__name__)
 
@@ -55,7 +65,7 @@ class KombuRPCClient(rpc_base.RPCClient, kombu_base.Base):
 
         hosts = self._hosts.get_hosts()
 
-        self._connections = []
+        connections = []
 
         for host in hosts:
             conn = self._make_connection(
@@ -65,9 +75,9 @@ class KombuRPCClient(rpc_base.RPCClient, kombu_base.Base):
                 host.password,
                 self.virtual_host
             )
-            self._connections.append(conn)
+            connections.append(conn)
 
-        self.conn = self._connections[0]
+        self._connections = itertools.cycle(connections)
 
         # Create exchange.
         exchange = self._make_exchange(
@@ -104,7 +114,9 @@ class KombuRPCClient(rpc_base.RPCClient, kombu_base.Base):
         try:
             return self._listener.get_result(correlation_id, self._timeout)
         except moves.queue.Empty:
-            raise exc.MistralException("RPC Request timeout")
+            raise exc.MistralException(
+                "RPC Request timeout, correlation_id = %s" % correlation_id
+            )
 
     def _call(self, ctx, method, target, async_=False, **kwargs):
         """Performs a remote call for the given method.
@@ -126,27 +138,26 @@ class KombuRPCClient(rpc_base.RPCClient, kombu_base.Base):
             'async': async_
         }
 
-        LOG.debug("Publish request: {0}".format(body))
+        LOG.debug("Publish request: %s", body)
 
         try:
             if not async_:
                 self._listener.add_listener(correlation_id)
 
             # Publish request.
-            with kombu.producers[self.conn].acquire(block=True) as producer:
-                producer.publish(
-                    body=body,
-                    exchange=self.exchange,
-                    routing_key=self.topic,
-                    reply_to=self.queue_name,
-                    correlation_id=correlation_id,
-                    delivery_mode=2
-                )
+            for retry_round in six.moves.range(EPIPE_RETRIES):
+                if self._publish_request(body, correlation_id):
+                    break
 
             # Start waiting for response.
             if async_:
                 return
 
+            LOG.debug(
+                "Waiting a reply for sync call [reply_to = %s]",
+                self.queue_name
+            )
+
             result = self._wait_for_result(correlation_id)
             res_type = result[kombu_base.TYPE]
             res_object = result[kombu_base.RESULT]
@@ -162,6 +173,39 @@ class KombuRPCClient(rpc_base.RPCClient, kombu_base.Base):
 
         return res_object
 
+    def _publish_request(self, body, correlation_id):
+        """Publishes the request message
+
+        .. note::
+            The :const:`errno.EPIPE` socket errors are suppressed
+            and result in False being returned. This is because
+            this type of error can usually be fixed by retrying.
+
+        :param body: message body
+        :param correlation_id: correlation id
+        :return: True if publish succeeded, False otherwise
+        :rtype: bool
+        """
+        try:
+            conn = self._listener.wait_ready()
+            if conn:
+                with kombu.producers[conn].acquire(block=True) as producer:
+                    producer.publish(
+                        body=body,
+                        exchange=self.exchange,
+                        routing_key=self.topic,
+                        reply_to=self.queue_name,
+                        correlation_id=correlation_id,
+                        delivery_mode=2
+                    )
+                    return True
+        except socket.error as e:
+            if e.errno != errno.EPIPE:
+                raise
+            else:
+                LOG.debug('Retrying publish due to broker connection failure')
+        return False
+
     def sync_call(self, ctx, method, target=None, **kwargs):
         return self._call(ctx, method, async_=False, target=target, **kwargs)
 
